@@ -17,7 +17,8 @@ use fields (
 	'auth',    # [ user,pass ] or { realm1 => [ user1,pass1 ], realm2 => [ user2,pass2 ],... }
 			   # or callback(realm,user)->pass
 			   # if given, handle_response might automatically try to authorize requests
-	'contact', # optional contact
+	'contact', # optional local contact
+	'remote_contact', # remote contact from response
 	'callid',  # call-id value
 	'cseq',    # number in cseq header
 	'route',   # for 'route' header, comes usually from 'record-route' info in response
@@ -168,12 +169,10 @@ sub new_request {
 
 		$method = uc($method);
 		my $uri = delete $args{uri};
-		my ($to,$from,$contact,$remote_contact) = $self->{incoming}
-			? ( $self->{from}, $self->{to},undef,$self->{contact} )
-			: ( $self->{to}, $self->{from},$self->{contact},undef )
-			;
+		my ($to,$from) = $self->{incoming} ? ($self->{from},$self->{to}) 
+			: ($self->{to},$self->{from});
 		if ( !$uri ) {
-			($uri) = sip_hdrval2parts( to => $remote_contact||$to);
+			($uri) = sip_hdrval2parts( to => $self->{remote_contact}||$to);
 			# XXX handle quotes right, e.g "<bla>" <sip:bla@fasel.com>
 			$uri = $1 if $uri =~m{<(\S+)>$};
 		}
@@ -187,7 +186,7 @@ sub new_request {
 			{
 				from => $from,
 				to => $to,
-				$contact ? ( contact => $contact ):(),
+				$self->{contact} ? ( contact => $self->{contact} ):(),
 				route => $self->{route},
 				cseq => "$cseq $method",
 				'call-id' => $self->{callid},
@@ -304,13 +303,16 @@ sub handle_response {
 	# Don't care about the response for a CANCEL  or a BYE
 	# because this connection close is issued by this side
 	# and no matter what the peer wants the call be will closed
-	# But invoke callback to notify upper layer that the BYE was received
+	# But invoke callback to notify upper layer
 	if ( $method eq 'CANCEL' or $method eq 'BYE' ) {
 		if ( $code >=100 and $code<=199 ) {
 			push @ntrans,$tr
 		} else {
 			invoke_callback($cb,@arg,0,$code,$response,$leg,$from);
-			$endpoint->close_context( $self );
+			# close context only for BYE,
+			# for CANCEL we will close the context on receiving the
+			# response and sending the ACK
+			$endpoint->close_context( $self ) if $method eq 'BYE';
 		}
 		return;
 	}
@@ -351,16 +353,11 @@ sub handle_response {
 			# and propagate to upper layer
 			my $req = $tr->{request};
 
-			# create ACK, the ACK needs to be send with the original URI,
-			# even if the contact header (see below) changes the dialogs URI
-			my $ack = $req->create_ack( $response );
-			invoke_callback($cb,@arg,0,$code,$response,$leg,$from,$ack);
-			$endpoint->new_request( $ack,$self,undef,undef,leg => $leg, dst_addr => $from );
-
 			# 12.1.2 - set URI for dialog to contact given in response which
 			# establishes the dialog
 			if ( my $contact = $response->get_header( 'contact' )) {
 				$contact = $1 if $contact =~m{<(\w+:[^>\s]+)>};
+				$self->{remote_contact} = $contact;
 				$req->set_uri( $contact );
 			}
 
@@ -368,7 +365,14 @@ sub handle_response {
 			# FIXME: this should probably be better done by the upper layer
 			# which decides, which call to accept (in case of call-forking with
 			# multiple 2xx responses)
-			$self->{to} = $response->get_header( 'to' );
+			$self->{to} = $response->get_header( 'to' ) if ! $self->{incoming};
+
+			# create ACK
+			# if 2xx response changed contact use it as the new URI
+			my $ack = $req->create_ack( $response );
+			invoke_callback($cb,@arg,0,$code,$response,$leg,$from,$ack);
+			$endpoint->new_request( $ack,$self,undef,undef,leg => $leg, dst_addr => $from );
+
 
 		} else {
 			# response to ACK, REGISTER...
@@ -403,6 +407,7 @@ sub handle_response {
 		# redo request and insert request again
 		my $contact = $self->{to} = $response->get_header( 'contact' );
 		$contact = $1 if $contact =~m{<(\w+:[^>\s]+)>};
+		$self->{remote_contact} = $contact;
 		( my $r = $tr->{request} )->set_uri( $contact );
 		$r->set_cseq( ++$self->{cseq} );
 		$endpoint->new_request( $r,$self );
@@ -507,6 +512,16 @@ sub handle_request {
 		# invoke callback before closing context, so that we have more
 		# information about the current call
 		invoke_callback($cb,@arg,0,0,$request,$leg,$from);
+
+		if ( $method eq 'CANCEL' ) {
+			# must create 487 Request canceled
+			my $response = $request->create_response( '487','Request canceled' );
+			$response->set_header(
+				cseq => $response->cseq =~m{(\d+)} && "$1 INVITE" );
+			DEBUG(10,"send response: ".$response->dump(1));
+			$endpoint->new_response($self,$response,$leg,$from);
+		}
+
 		$endpoint->close_context($self);
 		return;
 	}
